@@ -4,7 +4,7 @@ use warnings FATAL => 'all';
 
 package namespace::local;
 
-our $VERSION = '0.0603';
+our $VERSION = '0.0604';
 
 =head1 NAME
 
@@ -279,8 +279,7 @@ sub prepare {
     #     thus preventing subsequent imports from leaking upwards
     # I do not know why it works, it shouldn't.
     if ($action eq '-around') {
-        $self->erase_symbols;
-        $self->write_symbols( $table );
+        $self->replace_symbols( undef, $table );
     };
 
     if ($action eq '-above' ) {
@@ -289,8 +288,7 @@ sub prepare {
         };
     } else {
         $self->{todo} = sub {
-            $self->erase_symbols;
-            $self->write_symbols( $table );
+            $self->replace_symbols( undef, $table );
         };
     };
 };
@@ -306,6 +304,9 @@ sub execute {
         unless $self->{done}++;
 };
 
+# Don't touch NAME, PACKAGE, and GLOB itself
+my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
+
 # in: package name
 # out: sorted & filtered list of symbols
 
@@ -330,41 +331,23 @@ sub read_names {
     return @list;
 };
 
-# Don't touch NAME, PACKAGE, and GLOB itself
-my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
-
-# In: package
-# Out: (none)
-# Side effect: destroys symbol table
-sub erase_symbols {
-    my ($self, $list) = @_;
-
-    my $package = $self->{target};
-    $list ||= [ $self->read_names ];
-
-    foreach my $name( @$list ) {
-        next if $self->{touch_not}{$name};
-        no strict 'refs'; ## no critic
-        delete ${ $package."::" }{$name};
-    };
-};
-
 sub erase_only_symbols {
     my ($self, $table) = @_;
 
     my $package = $self->{target};
     my @list = keys %$table;
 
-    # read all necessary symbols
+    # load all necessary symbols
     my $current = $self->read_symbols( \@list );
 
+    # filter out what we were going to delete
     foreach my $name ( @list ) {
         $table->{$name}{$_} and delete $current->{$name}{$_}
             for @TYPES;
     };
 
-    $self->erase_symbols( \@list );
-    $self->write_symbols( $current );
+    # put it back in place
+    $self->replace_symbols( \@list, $current );
 };
 
 # In: package, symbol
@@ -389,19 +372,24 @@ sub read_symbols {
     return \%content;
 };
 
-# In: package, symbol, hash
-# Out: (none)
-# Side effect: recreates *package::symbol
-sub write_symbols {
-    my ($self, $table) = @_;
+# This method's signature is a bit counterintuitive:
+# $self->replace_symbols( \@names_to_erase, \%new_table_entries )
+# If first argument is omitted, the whole namespace is scanned instead.
+# Separate erase_symbols and write_symbols turned out to be cumbersome
+#    because of the need to handle granular exclusion list.
+# This method can fill both roles.
+# Providing an empty list would make it just write the symbol table,
+#    whereas an empty hash would mean deletion only.
+sub replace_symbols {
+    my ($self, $clear_list, $table) = @_;
 
     my $package = $self->{target};
 
-    # TODO lumping touch_not and $table together is a workaround
-    #     for erase_symbols skipping over touch_not altogether.
-    # It does not belong here, and should be fixed at some point.
+    $clear_list ||= [ $self->read_names ];
+    $table ||= {};
+
     my %uniq;
-    $uniq{$_}++ for keys %$table, keys %{ $self->{touch_not} };
+    $uniq{$_}++ for keys %$table, @$clear_list;
 
     foreach my $name( keys %uniq ) {
         my $copy = $table->{$name} || {};
@@ -420,18 +408,18 @@ sub write_symbols {
             delete ${ $package."::" }{$name};
         };
 
-        foreach my $type ( @TYPES ) {
-            defined $copy->{$type} or next;
+        foreach my $type ( keys %$copy ) {
+            ref $copy->{$type} or next;
             eval {
                 # FIXME on perls 5.014..5.022 this block fails
                 # because @ISA is readonly.
                 # So we wrap it in eval with no catch
                 # until a better solution is done
                 no strict 'refs'; ## no critic
-                *{ $package."::".$name } = $copy->{$type}
+                *{ $package."::".$name } = $copy->{$type};
+                1;
             } || do {
-                Carp::cluck "namespace::local: working around error: $@"
-                    unless $@ =~ /^Modification of a read-only/;
+                carp "namespace::local: failed to write $package :: $name ($type), but trying to continue: $@";
             };
         };
     };
